import abc
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    Generator,
    cast,
)

from pydantic import BaseModel, ValidationError, validator
from pydantic.utils import ValueItems, ROOT_KEY

IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]
DictStrAny = Dict[str, Any]


class AbstractChoiceTypeMixin(abc.ABC, BaseModel):
    @property
    def choice_type_fields(self) -> Set[str]:
        return set()

    @property
    def polymorphic_fields(self) -> Set[str]:
        return set()

    def _iter(
        self,
        to_dict: bool = False,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_choice_type: bool = False,
        exclude_polymorphic: bool = True,
    ) -> Generator[Tuple[str, Any], None, None]:
        if exclude_choice_type:
            exclude = ValueItems.merge(exclude, self.choice_type_fields)
        if exclude_polymorphic:
            exclude = ValueItems.merge(exclude, self.polymorphic_fields)
        return super()._iter(
            to_dict,
            by_alias,
            include,
            exclude,
            exclude_unset,
            exclude_defaults,
            exclude_none,
        )

    def dict(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_choice_type: bool = False,
        exclude_polymorphic: bool = True,
    ) -> DictStrAny:
        if exclude_choice_type:
            exclude = ValueItems.merge(exclude, self.choice_type_fields)
        if exclude_polymorphic:
            exclude = ValueItems.merge(exclude, self.polymorphic_fields)
        return dict(
            self._iter(
                to_dict=True,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )
        )

    def json(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_choice_type: bool = False,
        exclude_polymorphic: bool = True,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        encoder = cast(Callable[[Any], Any], encoder or self.__json_encoder__)

        # We don't directly call `self.dict()`, which does exactly this with `to_dict=True`
        # because we want to be able to keep raw `BaseModel` instances and not as `dict`.
        # This allows users to write custom JSON encoders for given `BaseModel` classes.
        data = dict(
            self._iter(
                to_dict=models_as_dict,
                by_alias=by_alias,
                include=include,
                exclude=exclude,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                exclude_choice_type=exclude_choice_type,
                exclude_polymorphic=exclude_polymorphic,
            )
        )
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        return self.__config__.json_dumps(data, default=encoder, **dumps_kwargs)


class ChoiceTypeMixinBase(AbstractChoiceTypeMixin):
    @property
    @abc.abstractmethod
    def _choice_type_fields(self) -> Set[str]:
        pass

    @property
    @abc.abstractmethod
    def _polymorphic_field(self) -> Set[str]:
        pass

    @property
    def choice_type_fields(self) -> Set[str]:
        fields = super().choice_type_fields.union(
            [
                f
                for f in self.choice_type_fields
                if hasattr(self, f) and getattr(self, f) is not None
            ]
        )
        return fields

    @property
    def polymorphic_fields(self) -> Set[str]:
        fields = super().polymorphic_fields.union([self._polymorphic_field])
        return fields


@validator("")
def validate_choice_types(cls, v, values):
    if v is not None:
        return v

    non_null_values = list(
        filter(
            lambda t: t[0] in cls._choice_type_fields and t[1] is not None,
            values.items(),
        )
    )
    if len(non_null_values) == 0:
        raise ValidationError(
            f"{cls.__class__.__name__}.{cls._polymorphic_field}[x] can not be None."
        )
    elif len(non_null_values) > 1:
        raise ValidationError(
            f"{cls.__class__.__name__}.{cls._polymorphic_field}[x] can only have one value."
        )
    return non_null_values[0][1]



from __future__ import annotations
from datetime import datetime
import itertools
from typing import (
    Any,
    ClassVar,
    ForwardRef,
    Literal,
    Optional,
    Sequence,
    Set,
    Union,
)
from pydantic import AnyUrl, BaseModel, Field, validator
from fhirkit.ChoiceTypeMixin import ChoiceTypeMixinBase, validate_choice_types
from fhirkit.data_types import Code, XHTML


class Element(BaseModel):
    id: Optional[str]
    extension: Sequence[Extension] = Field([], repr=False)


class Narrative(Element):
    status: Literal["generated", "extensions", "additional", "required"]
    div: XHTML


class Extension(Element):
    url: AnyUrl
    value: Optional[Any]


class BackboneElement(Element):
    modifierExtension: Sequence[Extension] = Field([], repr=False)


BackboneElement.update_forward_refs()
Element.update_forward_refs()


class AbstractCoding(Element):
    """FHIR Terminology based model for concepts"""

    display: Optional[str]
    code: str
    system: str
    version: Optional[str]

    class Config:
        allow_mutation = False


class Coding(AbstractCoding):
    def __repr__(self) -> str:
        if self.display:
            return f'"{self.display}" {self.system}|{self.code}'
        else:
            return f"{self.system}|{self.code}"

    def __str__(self) -> str:
        return self.display or f"{self.system}|{self.code}"

    def fsh(self, include_display: bool = False, include_version: bool = True):
        token = self.system
        if self.version is not None and include_version:
            token += "|" + self.version
        token += "#" + self.code
        if self.display is not None and include_display:
            token += ' "{self.display}"'
        return token

    def __eq__(self, other: AbstractCoding) -> bool:
        if isinstance(other, AbstractCoding):
            return self.system == other.system and self.code == other.code
        return False

    def __hash__(self):
        return hash((self.system, self.code))


CodeableConcept = ForwardRef("CodeableConcept")


class CodeableConcept(BaseModel):
    """FHIR Terminology based model for CodeableConcepts"""

    text: Optional[str]
    coding: Sequence[Coding] = Field(default=[])
    active: Optional[bool]

    def __str__(self) -> str:
        return self.text

    def __eq__(self, other: Union[CodeableConcept, Coding]) -> bool:

        if isinstance(other, Coding):
            result = any(other == c for c in self.coding)
            return result
        elif isinstance(other, CodeableConcept):
            return any(
                c1 == c2 for c1, c2 in itertools.product(self.coding, other.coding)
            )
        else:
            return False


CodeableConcept.update_forward_refs()


class Period(Element):
    start: Optional[datetime]
    end: Optional[datetime]


class Reference(Element):
    reference: Optional[str]
    type: Optional[AnyUrl]
    identifier: Optional[Identifier]
    display: Optional[str]


class Identifier(Element):
    use: Optional[Code]
    type: Optional[CodeableConcept]
    system: Optional[AnyUrl]
    value: Optional[str]
    period: Optional[Period]
    assigner: Optional[Reference]


Identifier.update_forward_refs()
Reference.update_forward_refs()


class Quantity(Element):
    value: float
    comparator: Optional[Literal["<", "<=", ">=", ">"]]
    unit: str
    system: AnyUrl
    code: Code

    def __str__(self) -> str:
        return f"{self.value} {self.unit}"

    def unit_as_coding(self):
        return Coding(display=self.unit, code=self.code, system=self.system)


class SimpleQuantity(Quantity):
    comparator: None = Field(..., const=True)


class Range(Element):
    low: Optional[SimpleQuantity]
    high: Optional[SimpleQuantity]


class ContactPoint(Element):
    system: Optional[
        Literal["phone", "fax", "email", "pager", "url", "sms", "other"]
    ] = None
    value: Optional[str] = None
    use: Optional[Literal["home", "work", "temp", "old", "mobile"]] = None
    period: Period


class UsageContextValueChoiceType(ChoiceTypeMixinBase):
    _choice_type_fields: ClassVar[Set[str]] = [
        "valueQuantity",
        "valueRange",
        "valueCodeableConcept",
        "valueReference",
    ]
    _polymorphic_field: ClassVar[Set[str]] = "value"
    valueQuantity: Optional[Quantity] = None
    valueRange: Optional[Range] = None
    valueCodeableConcept: Optional[CodeableConcept] = None
    valueReference: Optional[Reference] = None
    value: Union[Quantity, Range, CodeableConcept, Reference] = None

    validate_value = validator("value", pre=True, always=True, allow_reuse=True)(
        validate_choice_types
    )


class UsageContext(Element, UsageContextValueChoiceType):
    code: Coding



from importlib.resources import Resource
from typing import List, Literal, Optional

from pydantic import Field, HttpUrl
from elements import Identifier
from data_types import Instant
from fhirkit.elements import BackboneElement

BundleType = Literal[
    "document",
    "message",
    "transaction",
    "transaction-response",
    "batch",
    "batch-response",
    "history",
    "searchset",
    "collection",
]


class BundleLink(BackboneElement):
    relation: str
    url: HttpUrl


class BundleSearch(BackboneElement):
    # TODO
    pass


class BundleRequest(BackboneElement):
    method: Literal["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH"]
    url: HttpUrl


class BundleResponse(BackboneElement):
    status: str
    location: Optional[HttpUrl]
    etag: Optional[str]
    lastModified: Optional[Instant]
    outcome: Optional[Resource]


class BundleEntry(BackboneElement):
    fullUrl: Optional[HttpUrl]
    resource: Optional[Resource]
    request: Optional[BundleRequest]
    response: Optional[BundleResponse]


class Bundle(Resource):
    resourceType = Field("Bundle", const=True)
    identifier: Optional[Identifier]
    type: BundleType
    timestamp: Optional[Instant]
    entry: List[BundleEntry]



from .Resource import Resource
from .elements import CodeableConcept, Coding, Quantity, Reference
from .CodeSystem import CodeSystem
from .ValueSet import ValueSet, SimpleValueSet
from .snomed import SCTCoding, SCTConcept, SCTFHIRTerminologyServer
from .Observation import Observation
from .Procedure import Procedure
from .SimpleFHIRServer import SimpleFHIRServer



import abc
from typing import Optional, Tuple, Union
from pydantic.error_wrappers import ValidationError
from fhirkit.data_types import Code
from fhirkit.elements import CodeableConcept, Coding
from pydantic import HttpUrl, parse_obj_as
from abc import ABC


class ExpansionError(Exception):
    pass


class ResourceNoteFoundError(Exception):
    pass


class AbstractFHIRServer(ABC):
    def __init__(self, baseUrl: Union[str, HttpUrl]) -> None:
        self._base_url = parse_obj_as(HttpUrl, baseUrl)

    @property
    def base_url(self):
        return self._base_url

    @abc.abstractmethod
    def get_resource(
        self,
        type: Optional[str] = None,
        *,
        id: Optional[str] = None,
        url: Optional[str] = None,
    ):
        pass

    def __getitem__(self, key: Union[str, Tuple[str, str], HttpUrl]):
        # TODO implement URL keys
        if isinstance(key, HttpUrl):
            return self.get_resource(None, url=key)
        try:
            url = parse_obj_as(HttpUrl, key)
        except ValidationError:
            pass
        else:
            return self.get_resource(None, url=url)
        if isinstance(key, str):
            resourceType, id = key.split("/")
        elif isinstance(key, tuple):
            resourceType, id = key
        else:
            raise TypeError(f"Received a key(={key}) of type {type(key)}")
        return self.get_resource(resourceType, id=id)


class AbstractFHIRTerminologyServer(AbstractFHIRServer):
    @abc.abstractmethod
    def valueset_expand(self, *args, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def valueset_validate_code(
        self,
        url: HttpUrl,
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[HttpUrl] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        raise NotImplementedError()



from typing import Generic, Optional, Sequence, TypeVar, Union

from pydantic import HttpUrl
from fhirkit.Server import AbstractFHIRServer, ResourceNoteFoundError
from fhirkit.Resource import Resource

R = TypeVar("R", bound=Resource)


class SimpleFHIRServer(Generic[R], AbstractFHIRServer):
    def __init__(self, resources: Sequence[R], baseUrl: Union[str, HttpUrl]) -> None:
        self._resources = resources
        super().__init__(baseUrl)

    def get_resource(
        self,
        resourceType: str,
        *,
        id: Optional[str] = None,
        url: Optional[Union[str, HttpUrl]] = None,
    ):
        if url is None:
            if id is None:
                raise RuntimeError(
                    "At least resource.id or resource.url must be given to be able to identify the requested resource"
                )
            url = self.base_url + "/" + resourceType + "/" + id
        assert self._resources is not None
        for r in self._resources:
            if (r.resourceType == resourceType or resourceType is None) and (
                r.url == url or r.id == id
            ):
                return r
        t = resourceType or "Resource"
        raise ResourceNoteFoundError(f"{t} with url={url} not found.")



from typing import Literal, Optional, Union
from pydantic import BaseModel, Field, ValidationError, validator
from fhirkit.Resource import DomainResource
from fhirkit.elements import CodeableConcept, Identifier, Period, Reference
from fhirkit.data_types import dateTime


class ProcedurePerformedChoiceTypeMixin(BaseModel):
    performedDateTime: Optional[dateTime] = Field(None, exclude=True)
    performedPeriod: Optional[Period] = Field(None, exclude=True)
    performedString: Optional[str] = Field(None, exclude=True)
    performed: Union[dateTime, Period, str] = Field(None, repr=True)

    @validator("performed", pre=True, always=True, allow_reuse=True)
    def determine_performed(cls, v, values):
        if v is not None:
            return v

        # if Procedure.performed[x] is not send through 'value' check other typed field names
        non_null_values = list(
            filter(
                lambda t: t[0].startswith("performed") and t[1] is not None,
                values.items(),
            )
        )
        if len(non_null_values) == 0:
            raise ValidationError("Procedure.performed[x] can not be None.")
        elif len(non_null_values) > 1:
            raise ValidationError("Procedure.performed[x] can only have one value.")
        return non_null_values[0][1]


class Procedure(DomainResource, ProcedurePerformedChoiceTypeMixin):
    resourceType = Field("Procedure", const=True)
    identifier: Optional[Identifier] = Field(None, repr=True)
    status: Literal[
        "preparation",
        "in-progress",
        "not-done",
        "on-hold",
        "stopped",
        "completed",
        "entered-in-error",
        "unknown",
    ] = Field("completed", repr=True)
    statusReason: Optional[CodeableConcept] = Field(None, repr=True)
    category: Optional[CodeableConcept] = Field(None, repr=True)
    code: Optional[CodeableConcept] = Field(None, repr=True)
    subject: Reference
    encounter: Optional[Reference] = Field(None, repr=True)



from typing import Any, ClassVar, List, Optional, Set, Union
from pydantic import (
    Field,
    HttpUrl,
    StrictInt,
    StrictStr,
    StrictBool,
    validator,
)
from fhirkit.ChoiceTypeMixin import ChoiceTypeMixinBase, validate_choice_types
from fhirkit.Resource import Resource
from fhirkit.data_types import Code
from fhirkit.elements import BackboneElement, CodeableConcept, Coding


class ParameterValueChoiceTypeMixin(ChoiceTypeMixinBase):
    _choice_type_fields: ClassVar[Set[str]] = [
        "valueBoolean",
        "valueString",
        "valueCode",
        "valueCoding",
        "valueCodeableConcept",
        "valueUri",
        "valueInteger",
    ]
    _polymorphic_field: ClassVar[Set[str]] = "value"
    valueBoolean: Optional[StrictBool] = Field(None)
    valueString: Optional[StrictStr] = Field(None)
    valueCode: Optional[Code] = Field(None)
    valueCoding: Optional[Coding] = Field(None)
    valueCodeableConcept: Optional[CodeableConcept] = Field(None)
    valueUri: Optional[HttpUrl] = Field(None)
    valueInteger: Optional[StrictInt] = Field(None)
    value: Union[
        StrictBool, StrictStr, Code, Coding, CodeableConcept, HttpUrl, StrictInt
    ] = Field(None, exclude=True)

    validate_value = validator("value", pre=True, always=True, allow_reuse=True)(
        validate_choice_types
    )


class AbstractParameter(BackboneElement):
    name: str


class ValueParameter(
    AbstractParameter,
    ParameterValueChoiceTypeMixin,
):
    def __str__(self) -> str:
        return f"{self.name}:{self.value}"


class ResourceParameter(AbstractParameter):
    resource: Optional[Resource]


class MultiPartParameter(AbstractParameter):
    part: List[ValueParameter]

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.part:
                if param.name == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return "\n\tpart: \n" + "\n\t".join(" " + str(p) for p in self.part)


class Parameters(Resource):
    resourceType = Field("Parameters", const=True)
    parameter: List[Union[ValueParameter, ResourceParameter, MultiPartParameter]]

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.parameter:
                if param.name == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return "Parameters: \n" + "\n".join(" " + str(p) for p in self.parameter)



from __future__ import annotations
import logging
from urllib.parse import urlencode
import time
from typing import ClassVar, List, Optional, Union
from webbrowser import Opera
import requests
from pydantic import HttpUrl, ValidationError, parse_obj_as, parse_raw_as
from fhirkit.Parameter import Parameters
from fhirkit.Server import AbstractFHIRTerminologyServer
from fhirkit.ValueSet import VSCodingWithDesignation, VSExpansion, ValueSet
from fhirkit.data_types import Code
from fhirkit.elements import CodeableConcept, Coding
from fhirkit.OperationOutcome import OperationOutcome, OperationOutcomeException


# DEFAULT_SCT_URL = "https://browser.ihtsdotools.org/snowstorm/snomed-ct/fhir"
# DEFAULT_SCT_URL = "https://r4.ontoserver.csiro.au/fhir"
# TIRO_SCT_URL = "https://snowstorm-aovarw23xa-uc.a.run.app/fhir"


class ExpandedValueset(ValueSet):
    expansion: VSExpansion


Response = Union[Parameters, OperationOutcome]
VSExpansionResponse = Union[ExpandedValueset, OperationOutcome]


class SCTFHIRTerminologyServer(AbstractFHIRTerminologyServer):
    DEFAULT_URL: ClassVar[
        str
    ] = "https://browser.ihtsdotools.org/snowstorm/snomed-ct/fhir"
    DEFAULT_SERVER: ClassVar[Optional[SCTFHIRTerminologyServer]] = None
    RETRY_COUNT: int = 3
    RETRY_PAUSE: int = 10

    def __init__(self, baseUrl: Optional[Union[str, HttpUrl]] = None) -> None:
        baseUrl = baseUrl or self.DEFAULT_URL
        self.baseUrl = parse_obj_as(HttpUrl, baseUrl)

    @classmethod
    def default_server(cls):
        if cls.DEFAULT_SERVER is None:
            cls.DEFAULT_SERVER = SCTFHIRTerminologyServer()
        return cls.DEFAULT_SERVER

    def get_resource(
        self, resourceType: str, *, id: Optional[str] = None, url: Optional[str] = None
    ):
        path = resourceType
        if id:
            path += "/" + id
        req_url = f"{self.baseUrl}/{path}"
        if url:
            req_url = url
        headers = {"Accept": "application/json", "Content-type": "application/json"}
        try:
            raw_response = requests.get(req_url, headers=headers)

            if raw_response.status_code != 200:
                logging.warn(
                    "Received an error from the snowstorm server (%s) after sending the following request %s",
                    raw_response.text,
                    req_url,
                )
            response = parse_obj_as(Response, raw_response.json())
        except ValidationError:
            raise ConnectionError(
                f"Received a response that doesn't resemble a FHIR-server. Please check if the server at {self.baseUrl} is a valid FHIR-server"
            )
        except:
            raise RuntimeWarning(
                "Failed when calling {endpoint} on {baseurl}".format(
                    endpoint=path, baseurl=self.baseUrl
                )
            )

        if isinstance(response, OperationOutcome):
            raise OperationOutcomeException(response)

        return response

    def valueset_expand(self, url: Union[str, HttpUrl], **kwargs):
        path = "ValueSet/$expand"
        page_size = 200
        offset = 0

        more_results_available = True

        while more_results_available:
            query = "url={url}&count={count}&offset={offset}".format(
                url=url, count=page_size, offset=offset
            )
            query += "&" + urlencode(kwargs) if kwargs else ""
            req_url = f"{self.baseUrl}/{path}?{query}"
            headers = {
                "Accept": "application/json",
            }
            payload = ""

            try:

                raw_response = requests.request(
                    "GET", req_url, data=payload, headers=headers
                )

                if raw_response.status_code != 200:
                    logging.warn(
                        "Received an error from the snowstorm server (%s) after sending the following request %s",
                        raw_response.text,
                        req_url,
                    )
                response = parse_obj_as(VSExpansionResponse, raw_response.json())

                if isinstance(response, OperationOutcome):
                    raise OperationOutcomeException(response)

                yield response

                offset = response.expansion.offset + page_size
                remaining = max(response.expansion.total - offset, 0)

            except ValidationError:
                raise ConnectionError(
                    f"Received a response that doesn't resemble a FHIR-server. Please check if the server at {self.baseUrl} is a valid FHIR-server"
                )
            except:
                raise RuntimeWarning(
                    "Failed when calling {endpoint} on {baseurl}".format(
                        endpoint=path, baseurl=self.baseUrl
                    )
                )
            else:

                if remaining % (page_size * 10) == 0:
                    logging.debug(
                        f"{remaining} concepts remaining (total: {response.expansion.total} | page size: {page_size})."
                    )
                more_results_available = remaining > 0

    def valueset_validate_code(
        self,
        url: Union[str, HttpUrl],
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[HttpUrl] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        assert (
            code or coding or codeableConcept
        ), "At least a code, coding or codeableConcept must be given to validate."

        path = "ValueSet/$validate-code"

        parameters = [dict(name="url", valueUri=str(url))]

        if code and system:
            parameters.append(dict(name="code", valueCode=code))
            parameters.append(dict(name="system", valueUri=str(system)))
            if display:
                parameters.append(dict(name="display", valueString=display))
        if coding:
            parameters.append(dict(name="coding", valueCoding=coding))
        if codeableConcept:
            parameters.append(
                dict(name="codeableConcept", valueCodeableConcept=codeableConcept)
            )

        req_url = f"{self.baseUrl}/{path}"
        headersList = {"Accept": "application/json", "Content-type": "application/json"}
        payload = Parameters(parameter=parameters).json(exclude_none=True)
        try:

            raw_response = requests.request(
                "POST", req_url, data=payload, headers=headersList
            )
            if raw_response.status_code != 200:
                logging.warn(
                    "Received an error from the snowstorm server (%s) after sending the following request %s",
                    raw_response.text,
                    req_url,
                )
            response = parse_obj_as(Response, raw_response.json())
        except ValidationError:
            raise ConnectionError(
                f"Received a response that doesn't resemble a FHIR-server. Please check if the server at {self.baseUrl} is a valid FHIR-server"
            )
        except:
            raise RuntimeWarning(
                "Failed when calling {endpoint} on {baseurl}".format(
                    endpoint=path, baseurl=self.baseUrl
                )
            )

        if isinstance(response, OperationOutcome):
            raise OperationOutcomeException(response)

        return response


DEFAULT_TERMINOLOGY_SERVER = SCTFHIRTerminologyServer()



from __future__ import annotations
from typing import (
    AbstractSet,
    Any,
    List,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    Generator,
)
from pydantic import AnyUrl, BaseModel, HttpUrl, Field, PrivateAttr
from fhirkit.Server import AbstractFHIRTerminologyServer
from fhirkit.data_types import Code, Id, Instant
from fhirkit.elements import BackboneElement, Element, Narrative, Extension, Coding
from fhirkit.ChoiceTypeMixin import AbstractChoiceTypeMixin


class Meta(BaseModel):
    versionId: Optional[Id]
    lastUpdated: Optional[Instant]
    source: Optional[AnyUrl]
    profile: List[AnyUrl] = []
    security: List[Coding] = []
    tag: List[Coding] = []


class Resource(AbstractChoiceTypeMixin, BaseModel):
    _fhir_server: Optional[AbstractFHIRTerminologyServer] = PrivateAttr(None)

    resourceType: str
    id: Optional[str] = Field(None, repr=False)
    meta: Optional[Meta] = Field(None, repr=False)
    implicitRules: Optional[HttpUrl] = Field(None, repr=False)
    language: Optional[Code] = Field(None, repr=False)

    def record(
        self,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True,
        exclude_choice_type: bool = True,
        exclude_polymorphic: bool = False,
    ):
        tuple_generator = self._iter(
            False,
            by_alias,
            include,
            exclude,
            exclude_unset,
            exclude_defaults,
            exclude_none,
            exclude_choice_type=exclude_choice_type,
            exclude_polymorphic=exclude_polymorphic,
        )
        return dict(self._assemble_key_recursively(tuple_generator))

    def dict(
        self,
        *,
        by_alias: bool = False,
        include: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude: Union[
            AbstractSet[Union[int, str]], Mapping[Union[int, str], any]
        ] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        return dict(
            super()._iter(
                True,
                by_alias,
                include,
                exclude,
                exclude_unset,
                exclude_defaults,
                exclude_none,
                exclude_choice_type=False,
            )
        )

    def _assemble_key_recursively(
        self, obj: Union[Element, Resource]
    ) -> Generator[Tuple[Tuple[str, ...], Element]]:
        if isinstance(obj, (tuple, list)):
            for key, value in enumerate(obj):
                for childKeys, childValue in self._assemble_key_recursively(value):
                    yield (key, *childKeys), childValue
        elif isinstance(obj, (BackboneElement, Resource, Generator)):
            for key, value in obj:
                for childKeys, childValue in self._assemble_key_recursively(value):
                    yield (key, *childKeys), childValue
        else:
            yield (), obj

    @property
    def choice_type_fields(self) -> Set[str]:
        return set()

    @property
    def polymorphic_fields(self) -> Set[str]:
        return set()

    class Config:
        arbitrary_types_allowed = True


class DomainResource(Resource):
    text: Optional[Narrative] = Field(None, repr=False)
    contained: Sequence[Resource] = Field([], repr=False)
    extension: Sequence[Extension] = Field([], repr=False)
    modifierExtension: Sequence[Extension] = Field([], repr=False)



from datetime import datetime
from typing import ClassVar, Dict, Literal, Optional, Sequence, Set, Tuple, Union

from pydantic import BaseModel, Field, PrivateAttr, ValidationError, validator
from fhirkit.ChoiceTypeMixin import (
    AbstractChoiceTypeMixin,
    ChoiceTypeMixinBase,
    validate_choice_types,
)
from fhirkit.Resource import DomainResource
from fhirkit.data_types import Code, dateTime
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Quantity,
    Reference,
    BackboneElement,
)

# TODO find better solution for multitype value
class ObservationValueChoiceTypeMixin(ChoiceTypeMixinBase):
    _choice_type_fields: ClassVar[Set[str]] = [
        "valueString",
        "valueQuantity",
        "valueInteger",
        "valueCodeableConcept",
        "valueBoolean",
    ]
    _polymorphic_field: ClassVar[Set[str]] = "value"
    valueString: Optional[str] = None
    valueQuantity: Optional[Quantity] = None
    valueInteger: Optional[int] = None
    valueCodeableConcept: Optional[CodeableConcept] = None
    valueBoolean: Optional[bool] = None
    value: Union[str, Quantity, int, CodeableConcept, bool] = None

    validate_value = validator("value", pre=True, always=True, allow_reuse=True)(
        validate_choice_types
    )


class ObservationEffectiveChoiceTypeMixin(ChoiceTypeMixinBase):
    _choice_type_fields: ClassVar[Set[str]] = ["effectiveDateTime", "effectivePeriod"]
    _polymorphic_field: ClassVar[str] = "effective"
    effectiveDateTime: Optional[dateTime] = None
    effectivePeriod: Optional[Period] = None
    effective: Union[dateTime, Period] = None

    validate_effective = validator(
        "effective", pre=True, always=True, allow_reuse=True
    )(validate_choice_types)


class ObservationComponent(BackboneElement, ObservationValueChoiceTypeMixin):
    code: CodeableConcept


class Observation(
    DomainResource, ObservationValueChoiceTypeMixin, ObservationEffectiveChoiceTypeMixin
):

    resourceType = Field("Observation", const=True)
    identifier: Sequence[Identifier] = Field([], repr=True)
    status: Literal[
        "registered",
        "preliminary",
        "final",
        "amended",
        "corrected",
        "cancelled",
        "enterred-in-error",
        "unknown",
    ] = Field("final", repr=True)
    category: Optional[CodeableConcept] = Field(None, repr=True)
    code: CodeableConcept = Field(..., repr=True)
    subject: Optional[Reference]
    encounter: Optional[Reference]

    method: Optional[Code] = Field(None, repr=False)
    derivedFrom: Optional[Reference]

    component: Sequence[ObservationComponent] = []



from __future__ import annotations
from email.generator import Generator
from typing import Any, ClassVar, Literal, Optional, Sequence, Set, TypeVar, Union
from pydantic import AnyUrl, Field, StrictBool, StrictStr, validator
from fhirkit.Resource import DomainResource
from fhirkit.ChoiceTypeMixin import ChoiceTypeMixinBase, validate_choice_types
from fhirkit.elements import (
    BackboneElement,
    CodeableConcept,
    Coding,
    Identifier,
    UsageContext,
)
from fhirkit.data_types import Code, dateTime
from fhirkit.metadata_types import ContactDetail


class CodeLookupError(KeyError):
    pass


class CSConceptDesignation(BackboneElement):
    language: Optional[Code]
    use: Optional[Coding]
    value: str


class CSConceptPropertyValueChoiceTypeMixin(ChoiceTypeMixinBase):
    _choice_type_fields: ClassVar[Set[str]] = [
        "valueBoolean",
        "valueString",
        "valueCode",
        "valueCoding",
        "valueDateTime",
        "valueDecimal",
    ]
    _polymorphic_field: ClassVar[Set[str]] = "value"
    valueBoolean: Optional[StrictBool] = Field(None)
    valueString: Optional[StrictStr] = Field(None)
    valueCode: Optional[Code] = Field(None)
    valueCoding: Optional[Coding] = Field(None)
    valueDateTime: Optional[dateTime] = Field(None)
    valueDecimal: Optional[float] = Field(None)
    value: Union[StrictBool, StrictStr, Code, Coding, dateTime, float] = Field(
        None, exclude=True
    )

    validate_value = validator("value", pre=True, always=True, allow_reuse=True)(
        validate_choice_types
    )

    def __str__(self) -> str:
        return str(self.value)


class CSConceptProperty(BackboneElement, CSConceptPropertyValueChoiceTypeMixin):
    code: Code

    def __str__(self) -> str:
        return str(self.code) + ": " + super().__str__()


class CSConcept(BackboneElement):
    code: Code
    display: Optional[str]
    definition: Optional[str]
    designation: Sequence[CSConceptDesignation] = []
    property: Sequence[CSConceptProperty] = []
    concept: Sequence[CSConcept] = []

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.property:
                if param.code == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return (
            f"#{self.code} '{self.display}'"
            + "\nproperties: \n"
            + "\n".join(" " + str(p) for p in self.property if p.value is not None)
        )


class CSConceptLookup(CSConcept):
    name: str


C = TypeVar("C", bound=CSConcept)


def traverse_concepts(
    s: Sequence[C],
) -> Generator[C, None, None]:
    for c in s:
        yield c
        if len(c.concept) > 0:
            yield from traverse_concepts(c.concept)


class CodeSystem(DomainResource):
    resourceType = Field("CodeSystem", const=True)
    url: Optional[AnyUrl]
    identifier: Sequence[Identifier] = []
    version: Optional[str]
    name: Optional[str]
    title: Optional[str]
    status: Literal["draft", "active", "retired", "unknown"]
    experimental: Optional[bool]
    date: Optional[dateTime]
    publisher: Optional[str]
    contact: Sequence[ContactDetail] = []
    description: Optional[str]
    useContext: Sequence[UsageContext] = []
    jurisdiction: Sequence[CodeableConcept] = []
    purpose: Optional[str]
    copyright: Optional[str]
    caseSensitive: Optional[bool]
    valueSet: Optional[AnyUrl]
    hierarchyMeaning: Optional[
        Literal["grouped-by", "is-a", "part-of", "classified-with"]
    ]
    compositional: Optional[bool]
    versionNeeded: Optional[bool]
    content: Literal["not-present", "example", "fragment", "complete", "supplement"]
    supplements: Optional[AnyUrl]
    count: Optional[int]
    concept: Sequence[CSConcept] = []

    def lookup(
        self,
        code: Optional[Code] = None,
        coding: Optional[Coding] = None,
        date: Optional[dateTime] = None,
        displayLanguage: Optional[Code] = None,
        property: Optional[Code] = None,
    ):
        if len(self.concept) == 0:
            raise RuntimeWarning(
                "No strategy to lookup codes or concepts if concepts are not explicitly available under 'CodeSystem.concept'"
            )

        if self.name is None:
            raise RuntimeWarning("The CodeSystem has no name to use as display name")
        assert (
            code is not None or coding is not None
        ), "At least a code or coding is needed to lookup."
        for concept in traverse_concepts(self.concept):

            if (concept.code == code) or (
                coding
                and concept.code == coding.code
                and concept.display == coding.display
            ):
                return CSConceptLookup(
                    name=self.name,
                    **concept.dict(
                        include={
                            "code",
                            "designation",
                            "property",
                            "display",
                            "version",
                        },
                        exclude_unset=True,
                    ),
                )

        raise CodeLookupError(
            f"No concept found in CodeSystem for given code/coding. (code={code}, coding={coding})"
        )

    def __iter__(self):
        yield from traverse_concepts(self.concept)

    def __getitem__(self, key: Code | Coding) -> CSConceptLookup:
        if isinstance(key, (Code, str)):
            return self.lookup(code=key)
        elif isinstance(key, Coding):
            return self.lookup(coding=key)
        raise KeyError("Key should be a Code or a Coding but received " + key)

    CSConcept.update_forward_refs()



from typing import List, Literal, Optional, Tuple, Union
from pydantic import HttpUrl, PrivateAttr, root_validator
from fhirkit.ValueSet import VSCompose, VSExpansion, VSFilter, VSInclude, ValueSet
from fhirkit.elements import CodeableConcept, Coding
from fhirkit.snomed.consts import SCT_URI
from fhirkit.snomed.terminology import SCTFHIRTerminologyServer


class SCTDescendantsFilter(VSFilter):
    property: Literal["concept"]
    op: Literal["is-a"]
    value: str


class SCTECLFilter(VSFilter):
    property: Literal["constraint"]
    op: Literal["="]
    value: str


class SCTImplicitInclude(VSInclude):
    concept: List = []
    valueSetList = []
    filter: Tuple[Union[SCTDescendantsFilter, SCTECLFilter]]


class SCTImplicitCompose(VSCompose):
    include: Tuple[SCTImplicitInclude]


class SCTImplicitValueSet(ValueSet):
    _fhir_server: SCTFHIRTerminologyServer = PrivateAttr(
        SCTFHIRTerminologyServer.default_server()
    )
    url: Optional[Union[HttpUrl, str]]
    compose: Optional[SCTImplicitCompose]

    @root_validator
    def check_either_url_or_compose(cls, values):
        url, compose = values.get("url"), values.get("compose")
        if url is None and compose is None:
            raise ValueError(
                "Either a 'url' or a 'compose' section should be specified to have a valid ValueSet."
            )
        return values

    def ensure_fhir_server(self):
        if self._fhir_server is None:
            raise RuntimeWarning(
                "Can't expand an implicit SNOMED-CT without a FHIR server."
            )
        return True

    def equivalent_url(self):
        if self.url:
            return self.url
        else:
            assert (
                self.compose is not None
            ), "If no url is specified at least a compose section should be available"
            inclusion = self.compose.include[0]
            filter_rule = inclusion.filter[0]
            if filter_rule.property == "constraint" and filter_rule.op == "=":
                url = inclusion.version or inclusion.system or SCT_URI
                url += "?fhir_vs=ecl/" + inclusion.filter[0].value
            elif filter_rule.property == "concept" and filter_rule.op == "is-a":
                url = inclusion.version or inclusion.system or SCT_URI
                url += "?fhir_vs=isa/" + inclusion.filter[0].value
            else:
                raise RuntimeError(
                    f"Unexepected filter {inclusion.filter[0].json()} in inclusion criterium: {inclusion.json()}"
                )
            return url

    def expand(self, **kwargs):
        self.ensure_fhir_server()
        url = self.equivalent_url()
        for expanded_vs in self._fhir_server.valueset_expand(url, **kwargs):
            self.extend(expanded_vs.expansion.contains, extend_compose=False)

    def validate_code(self, code: Union[Coding, CodeableConcept]):
        self.ensure_fhir_server()
        url = self.equivalent_url()
        assert isinstance(
            code, (Coding, CodeableConcept)
        ), "code should be a Coding or CodeableConcept in order to be able to validate it against a ValueSet"
        if isinstance(code, Coding):
            response = self._fhir_server.valueset_validate_code(url, coding=code)
        else:
            response = self._fhir_server.valueset_validate_code(
                url, codeableConcept=code
            )
        return response.result
